#include "Geometry.h"
#include "Window.h"
#include "Fbo.h"
#include "FboManager.h"
#include "FullScreenQuad.h"
#include "DeferShader.h"


//Constructor
DeferShader::DeferShader(Geometry* geoLights_, std::string gp_fileName, std::string fp_fileName, std::string stateInfo_) {
    geoLights         = geoLights_;
    shaderProgFactory = ShaderProgFactory::getInstance();
    stateInfo         = stateInfo_;
    fullScreenQuad    = FullScreenQuad::getInstance();
    //Setup rendering to G-Buffer
    vp_ObjToEye    = shaderProgFactory->loadProgram("Source/ShaderPrograms/vp_ObjToEye.cg", "main", VERTEX_PROGRAM);
    fp_Gbuffer     = shaderProgFactory->loadProgram("Source/ShaderPrograms/DeferredShading/fp_Gbuffer.cg", "main", FRAGMENT_PROGRAM);
    cg_AlphaReject = cgGetNamedParameter(fp_Gbuffer, "alphaReject");
    cg_Material    = cgGetNamedParameter(fp_Gbuffer, "material");
    cg_TexColor    = cgGetNamedParameter(fp_Gbuffer, "texColor");
    cgGLSetParameter1f(cg_AlphaReject, 0.0);               //Initially no alpha testing
    cgGLSetParameter4f(cg_Material, 0.5, 1.0, 0.5, 16.0);  //Consists of ka, kd, ks, shininess
    fboManager     = FboManager::getInstance();
    window         = Window::getInstance();
    fbo            = fboManager->createFbo(0, 3, window->getWidth(), window->getHeight(), GL_TEXTURE_RECTANGLE_ARB, GL_NEAREST, GL_NEAREST, GL_RGBA16F_ARB, GL_RGBA);
    texSceneColor  = fbo->getTex(0);
    texScenePos    = fbo->getTex(1);
    texSceneNormal = fbo->getTex(2);
    //Setup ambient shading
    vp_NoTransform           = shaderProgFactory->loadProgram("Source/ShaderPrograms/vp_NoTransform.cg", "main", VERTEX_PROGRAM);
    fp_Ambient               = shaderProgFactory->loadProgram("Source/ShaderPrograms/DeferredShading/fp_Ambient.cg", "main", FRAGMENT_PROGRAM);
    cg_TexSceneColor_Ambient = cgGetNamedParameter(fp_Ambient, "texSceneColor");
    cgGLSetTextureParameter(cg_TexSceneColor_Ambient, texSceneColor);
    //Setup diffuse and specular shading
    gp_LightVolume    = shaderProgFactory->loadProgram(gp_fileName, "main", GEOMETRY_PROGRAM);
    fp_DeferShader    = shaderProgFactory->loadProgram(fp_fileName, "main", FRAGMENT_PROGRAM);
    cg_TexSceneColor  = cgGetNamedParameter(fp_DeferShader, "texSceneColor");
    cg_TexScenePos    = cgGetNamedParameter(fp_DeferShader, "texScenePos");
    cg_TexSceneNormal = cgGetNamedParameter(fp_DeferShader, "texSceneNormal");
    cgGLSetTextureParameter(cg_TexSceneColor, texSceneColor);
    cgGLSetTextureParameter(cg_TexScenePos, texScenePos);
    cgGLSetTextureParameter(cg_TexSceneNormal, texSceneNormal);
    //Add the contributions of each light using additive blending: out_col = 1*source_col + 1*dest_col
    glBlendFunc(GL_ONE, GL_ONE);
}


//drawAmbientScene
void DeferShader::drawAmbientScene() {
    //Texture color buffer to screen-sized quad and multiply it with the ambient term
    cgGLEnableTextureParameter(cg_TexSceneColor_Ambient);
    cgGLBindProgram(vp_NoTransform); cgGLBindProgram(fp_Ambient);
    cgGLEnableProfile(vertexProfile); cgGLEnableProfile(fragmentProfile);
    fullScreenQuad->draw();
    cgGLDisableTextureParameter(cg_TexSceneColor_Ambient);
}


//activate
void DeferShader::activate() {
    //Activate rendering to G-Buffer
    fbo->bind(3);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);  //Clear FBO
    cgGLBindProgram(vp_ObjToEye); cgGLBindProgram(fp_Gbuffer);
    cgGLEnableProfile(vertexProfile); cgGLEnableProfile(fragmentProfile);
}


//deactivate
void DeferShader::deactivate() {
    //Deactivate rendering to G-Buffer
    cgGLDisableProfile(vertexProfile); cgGLDisableProfile(fragmentProfile);
    fbo->unbind();

    //Perform deferred shading
    glDepthMask(GL_FALSE);
    //Draw ambient shaded scene
    drawAmbientScene();
    cgGLDisableProfile(fragmentProfile);
    //Enable additive blending and draw diffuse and specular scene (which is then added to the ambient shaded scene)
    glEnable(GL_BLEND);
    cgGLEnableTextureParameter(cg_TexSceneColor); cgGLEnableTextureParameter(cg_TexScenePos); cgGLEnableTextureParameter(cg_TexSceneNormal);
    cgGLBindProgram(gp_LightVolume); cgGLBindProgram(fp_DeferShader);  //vp_NoTransform is still bound
    cgGLEnableProfile(geometryProfile); cgGLEnableProfile(fragmentProfile);
    //Draw lights
    geoLights->draw(0, numLights);
    //Disable additive blending and shading
    cgGLDisableProfile(vertexProfile); cgGLDisableProfile(geometryProfile); cgGLDisableProfile(fragmentProfile);
    cgGLDisableTextureParameter(cg_TexSceneColor); cgGLDisableTextureParameter(cg_TexScenePos); cgGLDisableTextureParameter(cg_TexSceneNormal);
    glDisable(GL_BLEND);
    glDepthMask(GL_TRUE);
}


//enableTexSceneColor
void DeferShader::enableTexColor(GLuint texColor) {
    cgGLSetTextureParameter(cg_TexColor, texColor);
    cgGLEnableTextureParameter(cg_TexColor);
}


//disableTexSceneColor
void DeferShader::disableTexColor() {
    cgGLDisableTextureParameter(cg_TexColor);
}


//enableAlphaReject
void DeferShader::enableAlphaReject(GLfloat alphaReject) {
    cgGLSetParameter1f(cg_AlphaReject, alphaReject);
}


//disableAlphaReject
void DeferShader::disableAlphaReject() {
    cgGLSetParameter1f(cg_AlphaReject, 0.0);
}


//setNumLights
void DeferShader::setNumLights(GLushort numLights_) {
    numLights = numLights_;
}
