#include <cstdlib>
#include <iostream>
#include "Fbo.h"


//Constructor
Fbo::Fbo(GLsizei numAttachments, GLsizei fboWidth_, GLsizei fboHeight_) {
	fboWidth  = fboWidth_;
	fboHeight = fboHeight_;
	//Setup the FBO
	glGenFramebuffersEXT(1, &fbo);
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fbo);
	//Setup the render buffer for depth
	glGenRenderbuffersEXT(1, &depthBuffer);
	glBindRenderbufferEXT(GL_RENDERBUFFER_EXT, depthBuffer);
	glRenderbufferStorageEXT(GL_RENDERBUFFER_EXT, GL_DEPTH_COMPONENT, fboWidth, fboHeight);
	//And attach the depth render buffer to the FBO as it's depth attachment
	glFramebufferRenderbufferEXT(GL_FRAMEBUFFER_EXT, GL_DEPTH_ATTACHMENT_EXT, GL_RENDERBUFFER_EXT, depthBuffer);

    //Attach texture(s) to FBO
    textures = new GLuint(numAttachments);
    for (GLsizei i = 0; i < numAttachments; ++i) {
        glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fbo);  //Bind FBO
        //Setup texture
        glGenTextures(1, &textures[i]);
        glBindTexture(GL_TEXTURE_2D, textures[i]);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
        //NULL means reserve texture memory, but texels are undefined
        glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA16F_ARB, fboWidth, fboHeight, 0, GL_RGBA, GL_FLOAT, NULL);
        //And attach the texture to the FBO's color buffer so we can render to it
        glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT+i, GL_TEXTURE_2D, textures[i], 0);
        //Check if eveything went okay
        GLenum status = glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT);
        if (status != GL_FRAMEBUFFER_COMPLETE_EXT) {
            std::cerr << "Failed to create FBO\n";
            exit(-1);
        }
    }
    attachments = new GLenum(numAttachments);
    for (GLsizei i = 0; i < numAttachments; ++i) {
        attachments[i] = GL_COLOR_ATTACHMENT0_EXT+i;
    }

    //Unbind FBO for now (i.e. render to screen as usual)
    glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
}


//Constructor
Fbo::Fbo(GLsizei numAttachments, GLsizei fboWidth_, GLsizei fboHeight_, GLenum texType, GLenum minFilter, GLenum maxFilter, GLint internalFormat, GLenum format) {
    fboWidth  = fboWidth_;
    fboHeight = fboHeight_;
    //Setup the FBO
    glGenFramebuffersEXT(1, &fbo);
    glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fbo);
    //Setup the render buffer for depth
    glGenRenderbuffersEXT(1, &depthBuffer);
    glBindRenderbufferEXT(GL_RENDERBUFFER_EXT, depthBuffer);
    glRenderbufferStorageEXT(GL_RENDERBUFFER_EXT, GL_DEPTH_COMPONENT, fboWidth, fboHeight);
    //And attach the depth render buffer to the FBO as it's depth attachment
    glFramebufferRenderbufferEXT(GL_FRAMEBUFFER_EXT, GL_DEPTH_ATTACHMENT_EXT, GL_RENDERBUFFER_EXT, depthBuffer);

    //Attach texture(s) to FBO
    textures = new GLuint(numAttachments);
    for (GLsizei i = 0; i < numAttachments; ++i) {
        glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fbo);  //Bind FBO
        //Setup texture
        glGenTextures(1, &textures[i]);
        glBindTexture(texType, textures[i]);
        glTexParameteri(texType, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
        glTexParameteri(texType, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
        glTexParameteri(texType, GL_TEXTURE_MIN_FILTER, minFilter);
        glTexParameteri(texType, GL_TEXTURE_MAG_FILTER, maxFilter);
        //NULL means reserve texture memory, but texels are undefined
        glTexImage2D(texType, 0, internalFormat, fboWidth, fboHeight, 0, format, GL_FLOAT, NULL);
        //Generate mipmaps if required (must be before attaching it to FBO)
        if (minFilter != GL_NEAREST && minFilter != GL_LINEAR)
            glGenerateMipmapEXT(texType);
        //And attach the texture to the FBO's color buffer so we can render to it
        glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT+i, texType, textures[i], 0);
        //Check if eveything went okay
        GLenum status = glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT);
        if (status != GL_FRAMEBUFFER_COMPLETE_EXT) {
            std::cerr << "Failed to create FBO\n";
            exit(-1);
        }
    }
    attachments = new GLenum(numAttachments);
    for (GLsizei i = 0; i < numAttachments; ++i) {
        attachments[i] = GL_COLOR_ATTACHMENT0_EXT+i;
    }

    //Unbind FBO for now (i.e. render to screen as usual)
    glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
}


//Destructor
Fbo::~Fbo() {
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);  //Unbind the FBO
	glDeleteRenderbuffersEXT(1, &depthBuffer);
	glDeleteFramebuffersEXT(1, &fbo);
}


//getTex
GLuint Fbo::getTex(GLsizei attachment) {
    return textures[attachment];
}


//bind
void Fbo::bind(GLsizei numAttachments) {
	//Save the view port and set the new viewport to the size of the texture
	glPushAttrib(GL_VIEWPORT_BIT);
	glViewport(0, 0, fboWidth, fboHeight);
	//Bind the FBO so we can render to it
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fbo);
	//Inform OpenGL to which draw buffers we want to render
	glDrawBuffers(numAttachments, attachments);
}


//unbind
void Fbo::unbind() {
	//Restore the old previous view port (i.e. before rendering to the FBO)
	glPopAttrib();
	//Unbind the FBO (i.e. render to screen as usual)
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
}
