#include <cstdlib>
#include "Geometry.h"
#include <Soil/SOIL.h>


/* Classes for handling data for various geometry attrributes */
Geometry::GeometryAttribute::GeometryAttribute(){
    glGenBuffers(1, &vbo);
}


Geometry::GeometryAttribute::~GeometryAttribute() {
    glDeleteBuffers(1, &vbo);
}


void Geometry::PosAttribute::activate() {
    glBindBuffer(GL_ARRAY_BUFFER, vbo);
    glVertexPointer(numComponents, GL_FLOAT, 0, NULL);
    glEnableClientState(GL_VERTEX_ARRAY);
}


void Geometry::PosAttribute::deactivate() {
    glDisableClientState(GL_VERTEX_ARRAY);
}



void Geometry::ColorAttribute::activate() {
    glBindBuffer(GL_ARRAY_BUFFER, vbo);
    glColorPointer(numComponents, GL_FLOAT, 0, NULL);
    glEnableClientState(GL_COLOR_ARRAY);
}


void Geometry::ColorAttribute::deactivate() {
    glDisableClientState(GL_COLOR_ARRAY);
}


void Geometry::TexCoordAttribute::activate() {
    glBindBuffer(GL_ARRAY_BUFFER, vbo);
    glTexCoordPointer(numComponents, GL_FLOAT, 0, NULL);
    glEnableClientState(GL_TEXTURE_COORD_ARRAY);
}


void Geometry::TexCoordAttribute::deactivate() {
    glDisableClientState(GL_TEXTURE_COORD_ARRAY);
}


void Geometry::NormalAttribute::activate() {
    glBindBuffer(GL_ARRAY_BUFFER, vbo);
    glNormalPointer(GL_FLOAT, 0, NULL);  //Normals always consist of 3 components
    glEnableClientState(GL_NORMAL_ARRAY);
}


void Geometry::NormalAttribute::deactivate() {
    glDisableClientState(GL_NORMAL_ARRAY);
}


//Constructor
Geometry::Geometry(GLuint numVertices_, GLuint numIndices_) {
    numVertices = numVertices_;
    numIndices  = numIndices_;
    tex         = 0;
    init();
}


//Constructor
Geometry::Geometry(GLuint numVertices_, GLuint numIndices_, char const* fileNameTex) {
    numVertices = numVertices_;
    numIndices  = numIndices_;
    tex         = SOIL_load_OGL_texture(fileNameTex, SOIL_LOAD_AUTO, SOIL_CREATE_NEW_ID, SOIL_FLAG_MIPMAPS | SOIL_FLAG_INVERT_Y);
    init();
}


//Constructor
Geometry::Geometry(GLuint numVertices_, GLuint numIndices_, GLuint tex_) {
    numVertices = numVertices_;
    numIndices  = numIndices_;
    tex         = tex_;
    init();

}


/* Geometry class */
//init
void Geometry::init() {
    vboIndex    = 0;
    numAttribs  = 0, posIdx = 4, colorIdx = 4, texCoordIdx = 4, normalIdx = 4;
    for (GLubyte i = 0; i < 5; ++i)
        geoAttribs[i] = NULL;
    //Default is position at (0,0,0), no rotation, and no scaling
    xPos      = 0.0;
    yPos      = 0.0;
    zPos      = 0.0;
    angle     = 0.0;
    xRotation = 0.0;
    yRotation = 1.0;
    zRotation = 0.0;
    xScale    = 1.0;
    yScale    = 1.0;
    zScale    = 1.0;
}


//Destructor
Geometry::~Geometry() {
    delete[] geoAttribs;
    if (glIsTexture(tex))
        glDeleteTextures(1, &tex);
}


//draw
void Geometry::draw(GLuint startElement, GLuint numElements) {
    //Perform geometric transforms
    glPushMatrix();
    glTranslatef(xPos, yPos, zPos);
    glRotatef(angle, xRotation, yRotation, zRotation);
    glScalef(xScale, yScale, zScale);

    //Activate the necessary geometry attributes
    for (GLubyte i = 0; i < numAttribs; ++i)
        geoAttribs[i]->activate();
    //Draw the vertices, which the indices point to. However, only draw the vertices starting from startElement
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, vboIndex);
    glDrawElements(mode, numElements, GL_UNSIGNED_INT, (char*)NULL + startElement*sizeof(GLuint));
    //Deactivate the necessary geometry attributes
    for (GLubyte i = 0; i < numAttribs; ++i)
        geoAttribs[i]->deactivate();

    glPopMatrix();
}


//draw
void Geometry::draw() {
    //Perform geometric transforms
    glPushMatrix();
    glTranslatef(xPos, yPos, zPos);
    glRotatef(angle, xRotation, yRotation, zRotation);
    glScalef(xScale, yScale, zScale);

    //Activate the necessary geometry attributes
    for (GLubyte i = 0; i < numAttribs; ++i)
        geoAttribs[i]->activate();
    //Draw the vertices, which the indices point to
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, vboIndex);
    glDrawRangeElements(mode, 0, numIndices, numIndices, GL_UNSIGNED_INT, NULL);
    //Deactivate the necessary geometry attributes
    for (GLubyte i = 0; i < numAttribs; ++i)
        geoAttribs[i]->deactivate();

    glPopMatrix();
}


//setPositions
void Geometry::setPositions(AttributeComponents numComponents, GLfloat const* positions_, GLenum usage) {
    //Create a PosAttribute if it does not exist yet and add it to array geoAttribs
    if (geoAttribs[posIdx] == NULL) {
        posIdx = numAttribs;
        geoAttribs[posIdx] = new PosAttribute();
        positionBuffer = &geoAttribs[posIdx]->vbo;
        ++numAttribs;
    }

    geoAttribs[posIdx]->numComponents = numComponents;
    //Put vertex positions in VBO
    glBindBuffer(GL_ARRAY_BUFFER, geoAttribs[posIdx]->vbo);
    glBufferData(GL_ARRAY_BUFFER, numVertices*numComponents*sizeof(GLfloat), positions_, usage);
}


//setColors
void Geometry::setColors(AttributeComponents numComponents, GLfloat const* colors_, GLenum usage) {
    //Create a ColorAttribute if it does not exist yet and add it to array geoAttribs
    if (geoAttribs[colorIdx] == NULL) {
        colorIdx = numAttribs;
        geoAttribs[colorIdx] = new ColorAttribute();
        colorBuffer = &geoAttribs[colorIdx]->vbo;
        ++numAttribs;
    }

    geoAttribs[colorIdx]->numComponents = numComponents;
    //Put vertex colors in VBO
    glBindBuffer(GL_ARRAY_BUFFER, geoAttribs[colorIdx]->vbo);
    glBufferData(GL_ARRAY_BUFFER, numVertices*numComponents*sizeof(GLfloat), colors_, usage);
}


//setTexCoords
void Geometry::setTexCoords(AttributeComponents numComponents, GLfloat const* texCoords_, GLenum usage) {
    //Create a TexCoordAttribute if it does not exist yet and add it to array geoAttribs
    if (geoAttribs[texCoordIdx] == NULL) {
        texCoordIdx = numAttribs;
        geoAttribs[texCoordIdx] = new TexCoordAttribute();
        texCoordBuffer = &geoAttribs[texCoordIdx]->vbo;
        ++numAttribs;
    }

    geoAttribs[texCoordIdx]->numComponents = numComponents;
    //Put vertex texture coordinates in VBO
    glBindBuffer(GL_ARRAY_BUFFER, geoAttribs[texCoordIdx]->vbo);
    glBufferData(GL_ARRAY_BUFFER, numVertices*numComponents*sizeof(GLfloat), texCoords_, usage);
}


//setNormals
void Geometry::setNormals(GLfloat const* normals_, GLenum usage) {
    //Create a NormalAttribute if it does not exist yet and add it to array geoAttribs
    if (geoAttribs[normalIdx] == NULL) {
        normalIdx = numAttribs;
        geoAttribs[normalIdx] = new NormalAttribute();
        normalBuffer = &geoAttribs[normalIdx]->vbo;
        ++numAttribs;
    }

    geoAttribs[normalIdx]->numComponents = XYZ;  //Normals always consist of 3 components
    //Put vertex texture coordinates in VBO
    glBindBuffer(GL_ARRAY_BUFFER, geoAttribs[normalIdx]->vbo);
    glBufferData(GL_ARRAY_BUFFER, numVertices*XYZ*sizeof(GLfloat), normals_, usage);
}


//setIndices
void Geometry::setIndices(GLenum mode_, GLuint const* indices_, GLenum usage) {
    mode = mode_;
    //Generate index VBO if it does not exist yet
    if (!glIsBuffer(vboIndex)) {
        glGenBuffers(1, &vboIndex);
        indexBuffer = &vboIndex;
    }

    //Put indices in VBO
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, vboIndex);
    glBufferData(GL_ELEMENT_ARRAY_BUFFER, numIndices*sizeof(GLuint), indices_, usage);
}


//setTex
void Geometry::setTex(char const* fileNameTex, GLuint soilFlags) {
    tex = SOIL_load_OGL_texture(fileNameTex, SOIL_LOAD_AUTO, SOIL_CREATE_NEW_ID, soilFlags);
}


//setTransform
void Geometry::setTransform(GLfloat xPos_, GLfloat yPos_, GLfloat zPos_, GLfloat angle_, GLfloat scale) {
    xPos   = xPos_;
    yPos   = yPos_;
    zPos   = zPos_;
    angle  = angle_;
    xScale = scale;
    yScale = scale;
    zScale = scale;
}


//setRotationAxis
void Geometry::setRotationAxis(GLfloat x, GLfloat y, GLfloat z) {
    xRotation = x;
    yRotation = y;
    zRotation = z;
}


//setScale
void Geometry::setScale(GLfloat x, GLfloat y, GLfloat z) {
    xScale = x;
    yScale = y;
    zScale = z;
}


//Set and get functions
void Geometry::setPosBuffer(GLuint vbo_) {
    geoAttribs[posIdx]->vbo = vbo_;
}
void Geometry::setTexCoordsBuffer(GLuint vbo_) {
    geoAttribs[texCoordIdx]->vbo = vbo_;
}
GLuint Geometry::getPosBuffer() {
    return geoAttribs[posIdx]->vbo;
}
GLuint Geometry::getTexCoordsBuffer() {
    return geoAttribs[texCoordIdx]->vbo;
}
