#ifndef GEOMETRY_H
#define GEOMETRY_H
/* Class for storing and drawing geometry. It uses OpenGL's vertex buffer objects */


#include <GL/glew.h>


//Components of a vertex attribute
enum AttributeComponents {
    XY   = 2,
    XYZ  = 3,
    XYZW = 4,
    RGB  = 3,
    RGBA = 4
};


class Geometry {
    /* Classes for handling data for various geometry attrributes */
    class GeometryAttribute {
        public:
            GLuint vbo;
            AttributeComponents numComponents;


        public:
            //Constructor
            GeometryAttribute();

            //Destructor
            ~GeometryAttribute();

            //Activate VBO capabilities for the corresponding attribute
            virtual void activate() = 0;

            //Deactivate VBO capabilities for the corresponding attribute
            virtual void deactivate() = 0;
    };


    class PosAttribute: public GeometryAttribute {
        public:
            //Constructor
            PosAttribute(): GeometryAttribute() {}

            //Activate VBO capabilities for position
            void activate();

            //Deactivate VBO capabilities for position
            void deactivate();
    };


    class ColorAttribute: public GeometryAttribute {
        public:
            //Constructor
            ColorAttribute(): GeometryAttribute() {}

            //Activate VBO capabilities for position
            void activate();

            //Deactivate VBO capabilities for position
            void deactivate();
    };


    class TexCoordAttribute: public GeometryAttribute {
        public:
            //Constructor
            TexCoordAttribute(): GeometryAttribute() {}

            //Activate VBO capabilities for position
            void activate();

            //Deactivate VBO capabilities for position
            void deactivate();
    };


    class NormalAttribute: public GeometryAttribute {
        public:
            //Constructor
            NormalAttribute(): GeometryAttribute() {}

            //Activate VBO capabilities for position
            void activate();

            //Deactivate VBO capabilities for position
            void deactivate();
    };


    public:
        GLuint tex;  //Associated texture
        GLuint *positionBuffer, *colorBuffer, *normalBuffer, *texCoordBuffer, *indexBuffer;


    private:
        GLuint numVertices;
        GLuint numIndices;
        GLuint vboIndex;  //VBO containing the indices, of the vertices to be drawn
        GLenum mode;      //Type of primive to draw
        GLubyte numAttribs, posIdx, colorIdx, normalIdx, texCoordIdx;
        GeometryAttribute* geoAttribs[5];  //Contains active geometry attributes
        GLfloat xPos, yPos, zPos, angle, xRotation, yRotation, zRotation, xScale, yScale, zScale;  //Geometric transformations performed on geometry


    public:
        //Constructors
        Geometry(GLuint numVertices_, GLuint numIndices_);
        Geometry(GLuint numVertices_, GLuint numIndices_, char const* fileNameTex);
        Geometry(GLuint numVertices_, GLuint numIndices_, GLuint tex);

        //Destructor. Delete texture and allocated GeometryAttributes and their VBOs
        ~Geometry();

        //Initialize variables (function is called in constructor)
        void init();

        //Draw geometry
        void draw();

        //Draw part of geometry. I.e. only draw the vertices starting from startElement
        void draw(GLuint startElement, GLuint numElements);

        //Specify the positions of the geometry
        void setPositions(AttributeComponents numComponents, GLfloat const* positions_, GLenum usage);

        //Specify the colors of the geometry
        void setColors(AttributeComponents numComponents, GLfloat const* colors_, GLenum usage);

        //Specify the texture coordinates of the geometry
        void setTexCoords(AttributeComponents numComponents, GLfloat const* texCoords_, GLenum usage);

        //Specify the normals of the geometry
        void setNormals(GLfloat const* normals_, GLenum usage);

        //Specify the indices, of the vertices to be drawn. Parameter 'usage' denotes where and how the data is being used
        void setIndices(GLenum mode_, GLuint const* indices_, GLenum usage);

        //Specify texture and texture mapping flags
        void setTex(char const* fileNameTex, GLuint soilFlags);

        //Set geometry's position, rotation around y-axis (in degrees), and scale
        void setTransform(GLfloat xPos_, GLfloat yPos_, GLfloat zPos_, GLfloat angle_, GLfloat scale);

        //Set the rotation axis (overrides the default: rotation around y-axis)
        void setRotationAxis(GLfloat x, GLfloat y, GLfloat z);

        //Set the scaling
        void setScale(GLfloat x, GLfloat y, GLfloat z);

        //Set and get functions for setting and getting VBOs
        void setPosBuffer(GLuint vbo_);
        void setTexCoordsBuffer(GLuint vbo_);
        GLuint getPosBuffer();
        GLuint getTexCoordsBuffer();
};

#endif
