#include <cmath>
#include <cstdlib>
#include "Geometry.h"
#include "Timer.h"
#include "TransformFeedback.h"
#include "ShadingState.h"
#include "ShadingStateFactory.h"
#include "TradPhongLocal.h"
#include <Soil/SOIL.h>
#include "LightTornados.h"


#ifndef M_PI
#   define M_PI 3.14159265
#endif

LightTornados* LightTornados::singleton = NULL;


//Constructor
LightTornados::LightTornados(GLushort maxLights_, GLushort numLights_, GLfloat outerRadius_, GLfloat innerRadius) {
    maxLights           = maxLights_;
    numLights           = numLights_;
    outerRadius         = outerRadius_;
    geoLights           = new Geometry(maxLights, maxLights);
    shadingStateFactory = new ShadingStateFactory(maxLights, geoLights);
    activeState         = TradPhongLocal::getPointer();  //Start with traditional Phong shading and point lights with local influence
    shaderProgFactory   = ShaderProgFactory::getInstance();
    vp_Tornado          = shaderProgFactory->loadProgram("Source/ShaderPrograms/vp_Tornado.cg", "main", VERTEX_PROGRAM);
    cg_TornadoPos       = cgGetNamedParameter(vp_Tornado, "tornadoPos");
    cg_InnerAngle       = cgGetNamedParameter(vp_Tornado, "i");
    cg_OuterAngle       = cgGetNamedParameter(vp_Tornado, "o");
    cg_AddAngle         = cgGetNamedParameter(vp_Tornado, "a");
    outerAngle          = 0.0;
    transformFeedback   = new TransformFeedback(maxLights, 1, GL_STREAM_DRAW);
    transformFeedback->enablePositions(4);

    //Setup light sources
    GLfloat lightPos[4*maxLights];  //x,y,z,w-object-space light positions. Support for max. 1000 lights
    GLfloat spotDir[4*maxLights];   //x,y,z-object-space directions of spot lights and random number in w-component
    GLuint indicesLights[maxLights];

    //Setup light positions
    GLfloat innerAngle;
    GLfloat r;
    //Tornado left: (-outerRadius, 0.0, 0.0)
    GLint start = 0, end = maxLights / 5;
    for (GLint i = start; i < end; ++i) {
        innerAngle      = (GLfloat)(rand()) / RAND_MAX * 2.0*M_PI;  //Between 0 and 2*PI
        r               = (GLfloat)(rand()) / RAND_MAX + 3.0;       //Between 3 and 4
        lightPos[4*i]   = innerRadius*cos(innerAngle) - outerRadius;
        lightPos[4*i+1] = (GLfloat)((5*i+i) % 46 + r);             //Height between 3 and 50
        lightPos[4*i+2] = innerRadius*sin(innerAngle);
        lightPos[4*i+3] = 1.0;

    }
    //Tornado right: (outerRadius, 0.0, 0.0)
    start = end; end = 2 * (maxLights / 5);
    for (GLint i = start; i < end; ++i) {
        innerAngle      = (GLfloat)(rand()) / RAND_MAX * 2.0*M_PI;  //Between 0 and 2*PI
        r               = (GLfloat)(rand()) / RAND_MAX + 3.0;       //Between 3 and 4
        lightPos[4*i]   = innerRadius*cos(innerAngle) + outerRadius;
        lightPos[4*i+1] = (GLfloat)((5*i+i) % 46 + r);             //Height between 3 and 50
        lightPos[4*i+2] = innerRadius*sin(innerAngle);
        lightPos[4*i+3] = 1.0;
    }
    //Tornado far: (0.0, 0.0, -outerRadius)
    start = end; end = 3 * (maxLights / 5);
    for (GLint i = start; i < end; ++i) {
        innerAngle      = (GLfloat)(rand()) / RAND_MAX * 2.0*M_PI;  //Between 0 and 2*PI
        r               = (GLfloat)(rand()) / RAND_MAX + 3.0;       //Between 3 and 4
        lightPos[4*i]   = innerRadius*cos(innerAngle);
        lightPos[4*i+1] = (GLfloat)((5*i+i) % 46 + r);             //Height between 3 and 50
        lightPos[4*i+2] = innerRadius*sin(innerAngle) - outerRadius;
        lightPos[4*i+3] = 1.0;
    }
    //Tornado close: (0.0, 0.0, outerRadius)
    start = end; end = 4 * (maxLights / 5);
    for (GLint i = start; i < end; ++i) {
        innerAngle      = (GLfloat)(rand()) / RAND_MAX * 2.0*M_PI;  //Between 0 and 2*PI
        r               = (GLfloat)(rand()) / RAND_MAX + 3.0;       //Between 3 and 4
        lightPos[4*i]   = innerRadius*cos(innerAngle);
        lightPos[4*i+1] = (GLfloat)((5*i+i) % 46 + r);             //Height between 3 and 50
        lightPos[4*i+2] = innerRadius*sin(innerAngle) + outerRadius;
        lightPos[4*i+3] = 1.0;
    }
    //Tornado center: (0.0, 0.0, outerRadius*0.25)
    start = end; end = maxLights;
    for (GLint i = start; i < end; ++i) {
        innerAngle      = (GLfloat)(rand()) / RAND_MAX * 2.0*M_PI;  //Between 0 and 2*PI
        r               = (GLfloat)(rand()) / RAND_MAX + 3.0;       //Between 3 and 4
        lightPos[4*i]   = innerRadius*cos(innerAngle);
        lightPos[4*i+1] = (GLfloat)((5*i+i) % 46 + r);             //Height between 3 and 50
        lightPos[4*i+2] = innerRadius*sin(innerAngle) + outerRadius*0.25;
        lightPos[4*i+3] = 1.0;
    }

    //Setup spot directions
    for (GLushort i = 0; i < maxLights; ++i) {
        spotDir[4*i]   =  0.0; //((GLfloat)(rand()) / RAND_MAX * 2.0 - 1.0) * 0.3;  //Between -0.3 and 0.3
        spotDir[4*i+1] = -1.0; //-0.7;
        spotDir[4*i+2] =  0.0; //((GLfloat)(rand()) / RAND_MAX * 2.0 - 1.0) * 0.3;  //Between -0.3 and 0.3
        spotDir[4*i+3] = (GLfloat)(rand()) / RAND_MAX * 0.03;               //Between 0 and 0.03
        //Normalize spot direction
        GLdouble distSq = spotDir[4*i]*spotDir[4*i] + spotDir[4*i+1]*spotDir[4*i+1] + spotDir[4*i+2]*spotDir[4*i+2];
        GLdouble factor = 1.0 / sqrt(distSq);
        spotDir[4*i]   *= factor;
        spotDir[4*i+1] *= factor;
        spotDir[4*i+2] *= factor;
    }

    //Setup indices for the geoLights
    for (GLuint i = 0; i < maxLights; ++i) {
        indicesLights[i] = i;
    }

    //Store light source data in geometry object
    geoLights->setPositions(XYZW, lightPos, GL_STATIC_DRAW);
    geoLights->setTexCoords(XYZW, spotDir, GL_STATIC_DRAW);
    geoLights->setIndices(GL_POINTS, indicesLights, GL_STATIC_DRAW);

    //Setup point sprites (light sources). Their texture, size, and how to blend
    texLightSource = SOIL_load_OGL_texture("Source/particle.png", 0, 0, SOIL_FLAG_POWER_OF_TWO | SOIL_FLAG_MIPMAPS | SOIL_FLAG_INVERT_Y);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexEnvi(GL_POINT_SPRITE, GL_COORD_REPLACE, GL_TRUE);       //Replace tex coords of point sprites with proper tex coords
    glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);  //Texture replaces color
    glPointSize(15.0);            //Set size
    glBlendFunc(GL_ONE, GL_ONE);  //Set additive blending as blend function: out_col = 1*source_col + 1*dest_col

    //Setup timing
    timer = Timer::getInstance();
    tPrev = timer->time;  //Starting time
}


//getInstance
LightTornados* LightTornados::getInstance(GLushort maxLights_, GLushort numLights_, GLfloat outerRadius, GLfloat innerRadius) {
    if (singleton == NULL)
        singleton = new LightTornados(maxLights_, numLights_, outerRadius, innerRadius);
    return singleton;
}


//getPointer
LightTornados* LightTornados::getPointer() {
    return singleton;
}


//tornadoAnimation
void LightTornados::tornadoAnimation() {
    GLushort start = 0, numElements = maxLights / 5;
    //Tornado animation based on elapsed time, to prevent speed-ups/slow-downs on fast/slow hardware
    double timing = timer->time - tPrev;
    tPrev = timer->time;

    transformFeedback->bind(GL_POINTS);  //Render points to output buffer

    //Activate tornado animation
    addAngle = 0.3*timing;
    outerAngle += addAngle;
    cgGLSetParameter1f(cg_InnerAngle, 10.0*timing);
    cgGLSetParameter1f(cg_AddAngle, addAngle);
    cgGLSetParameter1f(cg_OuterAngle, outerAngle);
    cgGLBindProgram(vp_Tornado);
    cgGLEnableProfile(vertexProfile);

    //Left
    cgGLSetParameter3f(cg_TornadoPos, -outerRadius, 0.0, 0.0);
    geoLights->draw(start, numElements);
    //Right
    start += numElements;
    cgGLSetParameter3f(cg_TornadoPos, outerRadius, 0.0, 0.0);
    geoLights->draw(start, numElements);
    //Far
    start += numElements;
    cgGLSetParameter3f(cg_TornadoPos, 0.0, 0.0, -outerRadius);
    geoLights->draw(start, numElements);
    //Close
    start += numElements;
    cgGLSetParameter3f(cg_TornadoPos, 0.0, 0.0, outerRadius);
    geoLights->draw(start, numElements);
    //Center
    start += numElements;
    cgGLSetParameter3f(cg_TornadoPos, 0.0, 0.0, outerRadius*0.25);
    geoLights->draw(start, numElements);

    //Deactivate tornado animation
    cgGLDisableProfile(vertexProfile);
    transformFeedback->unbind();

    transformFeedback->swapBuffers(geoLights->positionBuffer, transformFeedback->positionBuffer);
}


//activate
void LightTornados::activate() {
    tornadoAnimation();
    activeState->setNumLights(numLights);
    activeState->activate();
}


//deactivate
void LightTornados::deactivate() {
    activeState->deactivate();
    drawLightSources();  //Draw alpha-blended light sources at the end
}


//drawLightSources
void LightTornados::drawLightSources() {
    glDepthMask(GL_FALSE);
    glEnable(GL_BLEND);
    glEnable(GL_TEXTURE_2D);
    glBindTexture(GL_TEXTURE_2D, texLightSource);
    glEnable(GL_POINT_SPRITE);

    geoLights->draw(0, numLights);

    glDisable(GL_POINT_SPRITE);
    glDisable(GL_TEXTURE_2D);
    glDisable(GL_BLEND);
    glDepthMask(GL_TRUE);
}


//addLights
void LightTornados::addLights(GLushort numAdded) {
    numLights += numAdded;
    if (numLights > maxLights)
        numLights = maxLights;

}


//deleteLights
void LightTornados::deleteLights(GLushort numDeleted) {
    if (numLights >= numDeleted)
        numLights -= numDeleted;
}
