#include <cstdlib>
#include <string>
#include <sstream>
#include "Eye.h"
#include "Timer.h"
#include "LightTornados.h"
#include "ShadingState.h"
#include "Demonstration.h"
#include <GL/glut.h>
#include "UserInterface.h"


static Eye* eye;
static LightTornados* lightTornados;
static Demonstration* demonstration;
static void (*pointerControls)();  //Function pointer to drawControls or drawNoControls


//drawBitmapString
void drawBitmapString(GLfloat x, GLfloat y, char const* text) {
    const char *c;
    glRasterPos2f(x, y);  //Set position to start drawing text
    //Iterate over all the characters in the string
    for (c = text; *c != '\0'; ++c) {
        glutBitmapCharacter(GLUT_BITMAP_HELVETICA_12, *c);
    }
}


//drawNoControls
void drawNoControls() {
    drawBitmapString( 0.74,-0.96, "Press [c] to show controls");
}


//drawControls
void drawControls() {
    drawBitmapString(-0.97,-0.26, "Keyboard:");
    drawBitmapString(-0.97,-0.31, "[d]\t\tDeferred shading");
    drawBitmapString(-0.97,-0.36, "[t]\t\t\tTraditional shading");
    drawBitmapString(-0.97,-0.41, "[s]\t\tSpot lights");
    drawBitmapString(-0.97,-0.46, "[l]\t\t\tPoint lights with local influence");
    drawBitmapString(-0.97,-0.51, "[g]\t\tPoint lights with global influence");
    drawBitmapString(-0.97,-0.56, "[1]\t\tPhong illumination");
    drawBitmapString(-0.97,-0.61, "[2]\t\tBlinn-Phong illumination");
    drawBitmapString(-0.97,-0.66, "[3]\t\tCook-Torrance illumination");
    drawBitmapString(-0.97,-0.71, "[4]\t\tGooch illumination");
    drawBitmapString(-0.97,-0.76, "[+]\t\tAdd a light");
    drawBitmapString(-0.97,-0.81, "[-]\t\tDelete a light");
    drawBitmapString(-0.97,-0.86, "[p]\t\tAdd an object");
    drawBitmapString(-0.97,-0.91, "[o]\t\tDelete an object");
    drawBitmapString(-0.97,-0.96, "[q]\t\tQuit");
    drawBitmapString(-0.50,-0.26, "Mouse:");
    drawBitmapString(-0.50,-0.31, "Zoom\t\t:\t\thold middle button and move mouse up/down");
    drawBitmapString(-0.50,-0.36, "Tilt\t\t\t\t\t\t:\t\thold right button and move mouse up/down");
    drawBitmapString(-0.50,-0.41, "Rotate\t:\t\thold left button and move mouse left/right");
    drawBitmapString( 0.74,-0.96, "Press [c] to hide controls");
}


//keyPressed
void keyPressed(unsigned char key, int x, int y) {
    //usleep(100);  //Avoid thrashing this procedure

    switch (key) {
        case 'q': case 27:  //27 is ASCII code of escape button
            exit(0);  //Clean exit
        break;
        case 'c':
            //Enable/disable showing controls
            if (pointerControls == &drawNoControls)
                pointerControls = &drawControls;
            else
                pointerControls = &drawNoControls;
        break;
        case 't':
            lightTornados->activeState = lightTornados->activeState->enableTraditional();
        break;
        case 'd':
            lightTornados->activeState = lightTornados->activeState->enableDeferred();
        break;
        case 's':
            lightTornados->activeState = lightTornados->activeState->enableSpot();
        break;
        case 'l':
            lightTornados->activeState = lightTornados->activeState->enableLocal();
        break;
        case 'g':
            lightTornados->activeState = lightTornados->activeState->enableGlobal();
        break;
        case '1':
            lightTornados->activeState = lightTornados->activeState->enablePhong();
        break;
        case '2':
            lightTornados->activeState = lightTornados->activeState->enableBlinnPhong();
        break;
        case '3':
            lightTornados->activeState = lightTornados->activeState->enableCookTorrance();
        break;
        case '4':
            lightTornados->activeState = lightTornados->activeState->enableGooch();
        break;
        case '+': case '=':
            lightTornados->addLights(1);
        break;
        case '-': case '_':
            lightTornados->deleteLights(1);
        break;
        case 'p':
            demonstration->addObject();
        break;
        case 'o':
            demonstration->deleteObject();
        break;
    }
    demonstration->enableUserInteraction();
}


//mouseButtonPressed
void mouseButtonPressed(int button, int state, int x, int y) {
    eye->mouseButtonPressed(button, state, x, y);
    demonstration->enableUserInteraction();
}


//mouseMotion
void mouseMotion(int x, int y) {
    eye->mouseMotion(x, y);
    demonstration->enableUserInteraction();
}


//mousePassiveMotion
void mousePassiveMotion(int x, int y) {
    demonstration->enableUserInteraction();
}


//Constructor
UserInterface::UserInterface(Demonstration* demonstration_) {
    demonstration     = demonstration_;
    eye               = Eye::getPointer();
    lightTornados     = LightTornados::getPointer();
    timer             = Timer::getInstance();
    shaderProgFactory = ShaderProgFactory::getInstance();
    vp_NoTransform    = shaderProgFactory->loadProgram("Source/ShaderPrograms/vp_NoTransform.cg", "main", VERTEX_PROGRAM);
    pointerControls   = &drawNoControls;

    glutKeyboardFunc(keyPressed);
    glutMouseFunc(mouseButtonPressed);
    glutMotionFunc(mouseMotion);
    glutPassiveMotionFunc(mousePassiveMotion);
}


//drawInfo
void UserInterface::drawInfo() {
    std::stringstream sstream;
    std::string infoFps, infoNumLights, infoSwitchTime;
    //Info fps
    timer->computeFps();
    sstream << (int)timer->fps;
    infoFps = "FPS: " + sstream.str();
    //Info number of lights
    sstream.str("");
    sstream << lightTornados->numLights;
    infoNumLights = sstream.str() + " lights";
    //Info switch time
    sstream.str("");
    sstream << demonstration->getSwitchTime();
    infoSwitchTime = "Switching in " + sstream.str() + " seconds";

    glColor3f(1.0, 1.0, 1.0);
    //No transforms on geometry to be processed
    cgGLBindProgram(vp_NoTransform);
    cgGLEnableProfile(vertexProfile);
    //Draw user info
    drawBitmapString(-0.97, 0.94, infoFps.c_str());
    drawBitmapString(-0.77, 0.94, infoNumLights.c_str());
    drawBitmapString(-0.59, 0.94, lightTornados->activeState->stateInfo.c_str());
    drawBitmapString( 0.75, 0.94, infoSwitchTime.c_str());
    //Draw with/without showing controls
    (*pointerControls)();
    //Deactive no transform on geometry
    cgGLDisableProfile(vertexProfile);
}
